@@ -1,21 +1,21 @@
 /*
-    Copyright 2005-2016 Intel Corporation.  All Rights Reserved.
+    Copyright (c) 2005-2017 Intel Corporation
+
+    Licensed under the Apache License, Version 2.0 (the "License");
+    you may not use this file except in compliance with the License.
+    You may obtain a copy of the License at
+
+        http://www.apache.org/licenses/LICENSE-2.0
+
+    Unless required by applicable law or agreed to in writing, software
+    distributed under the License is distributed on an "AS IS" BASIS,
+    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
+    See the License for the specific language governing permissions and
+    limitations under the License.
+
+
 
-    This file is part of Threading Building Blocks. Threading Building Blocks is free software;
-    you can redistribute it and/or modify it under the terms of the GNU General Public License
-    version 2  as  published  by  the  Free Software Foundation.  Threading Building Blocks is
-    distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
-    implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
-    See  the GNU General Public License for more details.   You should have received a copy of
-    the  GNU General Public License along with Threading Building Blocks; if not, write to the
-    Free Software Foundation, Inc.,  51 Franklin St,  Fifth Floor,  Boston,  MA 02110-1301 USA
 
-    As a special exception,  you may use this file  as part of a free software library without
-    restriction.  Specifically,  if other files instantiate templates  or use macros or inline
-    functions from this file, or you compile this file and link it with other files to produce
-    an executable,  this file does not by itself cause the resulting executable to be covered
-    by the GNU General Public License. This exception does not however invalidate any other
-    reasons why the executable file might be covered by the GNU General Public License.
 */
 
 #ifndef __TBB_partitioner_H
@@ -35,7 +35,7 @@
 #endif
 #ifndef __TBB_DEMAND_DEPTH_ADD
 // when imbalance is found range splits this value times more
-#define __TBB_DEMAND_DEPTH_ADD 2
+#define __TBB_DEMAND_DEPTH_ADD 1
 #endif
 #ifndef __TBB_STATIC_THRESHOLD
 // necessary number of clocks for the work to be distributed among all tasks
@@ -64,9 +64,7 @@ namespace tbb {
 
 class auto_partitioner;
 class simple_partitioner;
-#if TBB_PREVIEW_STATIC_PARTITIONER
 class static_partitioner;
-#endif
 class affinity_partitioner;
 
 namespace interface9 {
@@ -130,7 +128,7 @@ class flag_task: public task {
 public:
     tbb::atomic<bool> my_child_stolen;
     flag_task() { my_child_stolen = false; }
-    task* execute() { return NULL; }
+    task* execute() __TBB_override { return NULL; }
     static void mark_task_stolen(task &t) {
         tbb::atomic<bool> &flag = static_cast<flag_task*>(t.parent())->my_child_stolen;
 #if TBB_USE_THREADING_TOOLS
@@ -258,11 +256,29 @@ struct partition_type_base {
     }
 };
 
+//! Class determines whether template parameter has static boolean constant
+//! 'is_splittable_in_proportion' initialized with value of 'true' or not.
+/** If template parameter has such field that has been initialized with non-zero
+*  value then class field will be set to 'true', otherwise - 'false'
+*/
+template <typename Range>
+class is_splittable_in_proportion {
+private:
+    typedef char yes[1];
+    typedef char no[2];
+
+    template <typename range_type> static yes& decide(typename enable_if<range_type::is_splittable_in_proportion>::type *);
+    template <typename range_type> static no& decide(...);
+public:
+    // equals to 'true' if and only if static const variable 'is_splittable_in_proportion' of template parameter
+    // initialized with the value of 'true'
+    static const bool value = (sizeof(decide<Range>(0)) == sizeof(yes));
+};
+
 //! Provides default splitting strategy for partition objects.
 template <typename Partition>
 struct adaptive_mode : partition_type_base<Partition> {
     typedef Partition my_partition;
-    using partition_type_base<Partition>::self; // CRTP helper to get access to derived classes
     size_t my_divisor;
     // For affinity_partitioner, my_divisor indicates the number of affinity array indices the task reserves.
     // A task which has only one index must produce the right split without reserved index in order to avoid
@@ -271,12 +287,22 @@ struct adaptive_mode : partition_type_base<Partition> {
     static const unsigned factor = 1;
     adaptive_mode() : my_divisor(tbb::internal::get_initial_auto_partitioner_divisor() / 4 * my_partition::factor) {}
     adaptive_mode(adaptive_mode &src, split) : my_divisor(do_split(src, split())) {}
-    adaptive_mode(adaptive_mode &src, const proportional_split& split_obj) : my_divisor(do_split(src, split_obj)) {}
     /*! Override do_split methods in order to specify splitting strategy */
     size_t do_split(adaptive_mode &src, split) {
         return src.my_divisor /= 2u;
     }
-    size_t do_split(adaptive_mode &src, const proportional_split& split_obj) {
+};
+
+//! Provides proportional splitting strategy for partition objects
+template <typename Partition>
+struct proportional_mode : adaptive_mode<Partition> {
+    typedef Partition my_partition;
+    using partition_type_base<Partition>::self; // CRTP helper to get access to derived classes
+
+    proportional_mode() : adaptive_mode<Partition>() {}
+    proportional_mode(proportional_mode &src, split) : adaptive_mode<Partition>(src, split()) {}
+    proportional_mode(proportional_mode &src, const proportional_split& split_obj) { self().my_divisor = do_split(src, split_obj); }
+    size_t do_split(proportional_mode &src, const proportional_split& split_obj) {
 #if __TBB_ENABLE_RANGE_FEEDBACK
         size_t portion = size_t(float(src.my_divisor) * float(split_obj.right())
                                 / float(split_obj.left() + split_obj.right()) + 0.5f);
@@ -295,53 +321,8 @@ struct adaptive_mode : partition_type_base<Partition> {
         return portion;
     }
     bool is_divisible() { // part of old should_execute_range()
-        return my_divisor > my_partition::factor;
+        return self().my_divisor > my_partition::factor;
     }
-};
-
-//! Provides default linear indexing of partitioner's sequence
-template <typename Partition>
-struct linear_affinity_mode : adaptive_mode<Partition> {
-    using adaptive_mode<Partition>::my_divisor;
-    size_t my_head;
-    using adaptive_mode<Partition>::self;
-    linear_affinity_mode() : adaptive_mode<Partition>(), my_head(0) {}
-    linear_affinity_mode(linear_affinity_mode &src, split) : adaptive_mode<Partition>(src, split())
-        , my_head(src.my_head + src.my_divisor) {}
-    linear_affinity_mode(linear_affinity_mode &src, const proportional_split& split_obj) : adaptive_mode<Partition>(src, split_obj)
-        , my_head(src.my_head + src.my_divisor) {}
-    void set_affinity( task &t ) {
-        if( my_divisor )
-            t.set_affinity( affinity_id(my_head) + 1 );
-    }
-};
-
-//! Class determines whether template parameter has static boolean constant
-//! 'is_splittable_in_proportion' initialized with value of 'true' or not.
-/** If template parameter has such field that has been initialized with non-zero
- *  value then class field will be set to 'true', otherwise - 'false'
- */
-template <typename Range>
-class is_splittable_in_proportion {
-private:
-    typedef char yes[1];
-    typedef char no [2];
-
-    template <typename range_type> static yes& decide(typename enable_if<range_type::is_splittable_in_proportion>::type *);
-    template <typename range_type> static no& decide(...);
-public:
-    // equals to 'true' if and only if static const variable 'is_splittable_in_proportion' of template parameter
-    // initialized with the value of 'true'
-    static const bool value = (sizeof(decide<Range>(0)) == sizeof(yes));
-};
-
-//! Provides default methods for non-balancing partition objects.
-template<class Mode>
-struct unbalancing_partition_type : Mode {
-    using Mode::self;
-    unbalancing_partition_type() : Mode() {}
-    unbalancing_partition_type(unbalancing_partition_type& p, split) : Mode(p, split()) {}
-    unbalancing_partition_type(unbalancing_partition_type& p, const proportional_split& split_obj) : Mode(p, split_obj) {}
 #if _MSC_VER && !defined(__INTEL_COMPILER)
     // Suppress "conditional expression is constant" warning.
     #pragma warning( push )
@@ -350,7 +331,7 @@ struct unbalancing_partition_type : Mode {
     template <typename Range>
     proportional_split get_split() {
         if (is_splittable_in_proportion<Range>::value) {
-            size_t size = self().my_divisor / Mode::my_partition::factor;
+            size_t size = self().my_divisor / my_partition::factor;
 #if __TBB_NONUNIFORM_TASK_CREATION
             size_t right = (size + 2) / 3;
 #else
@@ -367,9 +348,25 @@ struct unbalancing_partition_type : Mode {
 #endif // warning 4127 is back
 };
 
+//! Provides default linear indexing of partitioner's sequence
+template <typename Partition>
+struct linear_affinity_mode : proportional_mode<Partition> {
+    size_t my_head;
+    using proportional_mode<Partition>::self;
+    linear_affinity_mode() : proportional_mode<Partition>(), my_head(0) {}
+    linear_affinity_mode(linear_affinity_mode &src, split) : proportional_mode<Partition>(src, split())
+        , my_head(src.my_head + src.my_divisor) {}
+    linear_affinity_mode(linear_affinity_mode &src, const proportional_split& split_obj) : proportional_mode<Partition>(src, split_obj)
+        , my_head(src.my_head + src.my_divisor) {}
+    void set_affinity( task &t ) {
+        if( self().my_divisor )
+            t.set_affinity( affinity_id(my_head) + 1 );
+    }
+};
+
 /*! Determine work-balance phase implementing splitting & stealing actions */
 template<class Mode>
-struct balancing_partition_type : unbalancing_partition_type<Mode> {
+struct dynamic_grainsize_mode : Mode {
     using Mode::self;
 #ifdef __TBB_USE_MACHINE_TIME_STAMPS
     tbb::internal::machine_tsc_t my_dst_tsc;
@@ -381,21 +378,21 @@ struct balancing_partition_type : unbalancing_partition_type<Mode> {
     } my_delay;
     depth_t my_max_depth;
     static const unsigned range_pool_size = __TBB_RANGE_POOL_CAPACITY;
-    balancing_partition_type(): unbalancing_partition_type<Mode>()
+    dynamic_grainsize_mode(): Mode()
 #ifdef __TBB_USE_MACHINE_TIME_STAMPS
         , my_dst_tsc(0)
 #endif
         , my_delay(begin)
         , my_max_depth(__TBB_INIT_DEPTH) {}
-    balancing_partition_type(balancing_partition_type& p, split)
-        : unbalancing_partition_type<Mode>(p, split())
+    dynamic_grainsize_mode(dynamic_grainsize_mode& p, split)
+        : Mode(p, split())
 #ifdef __TBB_USE_MACHINE_TIME_STAMPS
         , my_dst_tsc(0)
 #endif
         , my_delay(pass)
         , my_max_depth(p.my_max_depth) {}
-    balancing_partition_type(balancing_partition_type& p, const proportional_split& split_obj)
-        : unbalancing_partition_type<Mode>(p, split_obj)
+    dynamic_grainsize_mode(dynamic_grainsize_mode& p, const proportional_split& split_obj)
+        : Mode(p, split_obj)
 #ifdef __TBB_USE_MACHINE_TIME_STAMPS
         , my_dst_tsc(0)
 #endif
@@ -480,14 +477,14 @@ struct balancing_partition_type : unbalancing_partition_type<Mode> {
     }
 };
 
-class auto_partition_type: public balancing_partition_type<adaptive_mode<auto_partition_type> > {
+class auto_partition_type: public dynamic_grainsize_mode<adaptive_mode<auto_partition_type> > {
 public:
     auto_partition_type( const auto_partitioner& )
-        : balancing_partition_type<adaptive_mode<auto_partition_type> >() {
+        : dynamic_grainsize_mode<adaptive_mode<auto_partition_type> >() {
         my_divisor *= __TBB_INITIAL_CHUNKS;
     }
     auto_partition_type( auto_partition_type& src, split)
-        : balancing_partition_type<adaptive_mode<auto_partition_type> >(src, split()) {}
+        : dynamic_grainsize_mode<adaptive_mode<auto_partition_type> >(src, split()) {}
     bool is_divisible() { // part of old should_execute_range()
         if( my_divisor > 1 ) return true;
         if( my_divisor && my_max_depth ) { // can split the task. TODO: on-stack flag instead
@@ -519,31 +516,25 @@ public:
     }
 };
 
-#if TBB_PREVIEW_STATIC_PARTITIONER
-#ifndef __TBB_STATIC_PARTITIONER_BASE_TYPE
-#define __TBB_STATIC_PARTITIONER_BASE_TYPE unbalancing_partition_type
-#endif
-class static_partition_type : public __TBB_STATIC_PARTITIONER_BASE_TYPE<linear_affinity_mode<static_partition_type> > {
+class static_partition_type : public linear_affinity_mode<static_partition_type> {
 public:
     typedef proportional_split split_type;
     static_partition_type( const static_partitioner& )
-        : __TBB_STATIC_PARTITIONER_BASE_TYPE<linear_affinity_mode<static_partition_type> >() {}
+        : linear_affinity_mode<static_partition_type>() {}
     static_partition_type( static_partition_type& p, split )
-        : __TBB_STATIC_PARTITIONER_BASE_TYPE<linear_affinity_mode<static_partition_type> >(p, split()) {}
+        : linear_affinity_mode<static_partition_type>(p, split()) {}
     static_partition_type( static_partition_type& p, const proportional_split& split_obj )
-        : __TBB_STATIC_PARTITIONER_BASE_TYPE<linear_affinity_mode<static_partition_type> >(p, split_obj) {}
+        : linear_affinity_mode<static_partition_type>(p, split_obj) {}
 };
-#undef __TBB_STATIC_PARTITIONER_BASE_TYPE
-#endif
 
-class affinity_partition_type : public balancing_partition_type<linear_affinity_mode<affinity_partition_type> > {
+class affinity_partition_type : public dynamic_grainsize_mode<linear_affinity_mode<affinity_partition_type> > {
     static const unsigned factor_power = 4; // TODO: get a unified formula based on number of computing units
     tbb::internal::affinity_id* my_array;
 public:
     static const unsigned factor = 1 << factor_power; // number of slots in affinity array per task
     typedef proportional_split split_type;
     affinity_partition_type( tbb::internal::affinity_partitioner_base_v3& ap )
-        : balancing_partition_type<linear_affinity_mode<affinity_partition_type> >() {
+        : dynamic_grainsize_mode<linear_affinity_mode<affinity_partition_type> >() {
         __TBB_ASSERT( (factor&(factor-1))==0, "factor must be power of two" );
         ap.resize(factor);
         my_array = ap.my_array;
@@ -551,10 +542,10 @@ public:
         __TBB_ASSERT( my_max_depth < __TBB_RANGE_POOL_CAPACITY, 0 );
     }
     affinity_partition_type(affinity_partition_type& p, split)
-        : balancing_partition_type<linear_affinity_mode<affinity_partition_type> >(p, split())
+        : dynamic_grainsize_mode<linear_affinity_mode<affinity_partition_type> >(p, split())
         , my_array(p.my_array) {}
     affinity_partition_type(affinity_partition_type& p, const proportional_split& split_obj)
-        : balancing_partition_type<linear_affinity_mode<affinity_partition_type> >(p, split_obj)
+        : dynamic_grainsize_mode<linear_affinity_mode<affinity_partition_type> >(p, split_obj)
         , my_array(p.my_array) {}
     void set_affinity( task &t ) {
         if( my_divisor ) {
@@ -641,7 +632,6 @@ private:
     typedef interface9::internal::auto_partition_type::split_type split_type;
 };
 
-#if TBB_PREVIEW_STATIC_PARTITIONER
 //! A static partitioner
 class static_partitioner {
 public:
@@ -659,7 +649,6 @@ private:
     // TODO: consider to make split_type public
     typedef interface9::internal::static_partition_type::split_type split_type;
 };
-#endif
 
 //! An affinity partitioner
 class affinity_partitioner: internal::affinity_partitioner_base_v3 {
